import numpy as np

from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted
from sklearn.preprocessing import normalize
from pyod.models.base import BaseDetector
from dictlearn import dictionary_learning, sparse_encode, _get_fit_handle


class DL(BaseDetector):
    def __init__(self,
                 contamination,
                 n_components=100,
                 n_nonzero_coefs=3,
                 fit_algorithm='aksvd',
                 transform_algorithm='omp',
                 max_iter=50,
                 D0=None,
                 random_state=None):
        self.D = None
        self.decision_scores_ = None

        self.contamination = contamination
        self.n_components = n_components
        self.n_nonzero_coefs = n_nonzero_coefs
        self.fit_algorithm = _get_fit_handle(fit_algorithm)
        self.transform_algorithm = transform_algorithm
        self.max_iter = max_iter
        self.D0 = D0
        self.random_state = random_state
        self.params = {}

    def fit(self, X, y=None):
        """Fit detector. y is ignored in unsupervised methods.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples.
        y : Ignored
            Not used, present for API consistency by convention.
        Returns
        -------
        self : object
            Fitted estimator.
        """
        # validate inputs X and y (optional)
        Y = check_array(X.T)
        self._set_n_classes(y)

        if self.D0 is None:
            self.D0 = normalize(self.random_state.randn(Y.shape[0], self.n_components), axis=0, norm='l2')

        D, _, rmse, _ = dictionary_learning(Y, self.D0, None, self.max_iter, self.fit_algorithm,
                                            self.transform_algorithm, self.n_nonzero_coefs, self.params)

        self.D = D
        self.decision_scores_ = rmse
        self._process_decision_scores()
        return self

    def decision_function(self, X):
        """Predict raw anomaly score of X using the fitted detector.
        The anomaly score of an input sample is computed based on different
        detector algorithms. For consistency, outliers are assigned with
        larger anomaly scores.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only
            if they are supported by the base estimator.
        Returns
        -------
        anomaly_scores : numpy array of shape (n_samples,)
            The anomaly score of the input samples.
        """
        Y = check_array(X.T)
        check_is_fitted(self, ['decision_scores_', 'threshold_', 'labels_'])
        X, _ = sparse_encode(Y, self.D, n_nonzero_coefs=self.n_nonzero_coefs)
        err = np.linalg.norm(Y - self.D @ X, axis=0)
        return err
