# -*- coding: utf-8 -*-
"""Benchmark of all implemented algorithms
"""
# # Author: Script by Yue Zhao <zhaoy@cmu.edu>, adapted by Denis C. Ilie-Ablachim <denis.ilie_ablachim@upb.ro> and Bogdan Dumitrescu <bogdan.dumitrescu@upb.ro> for "Angle-based Dictionary Learning for Outlier Detection"
# License: BSD 2 clause

from __future__ import division
from __future__ import print_function

import os
import sys
from time import time

# temporary solution for relative imports in case pyod is not installed
# if pyod is installed, no need to use the following line
sys.path.append(
    os.path.abspath(os.path.join(os.path.dirname("__file__"), '..')))
# supress warnings for clean output
import warnings

warnings.filterwarnings("ignore")

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from scipy.io import loadmat

from pyod.models.abod import ABOD
from pyod.models.cblof import CBLOF
from pyod.models.feature_bagging import FeatureBagging
from pyod.models.hbos import HBOS
from pyod.models.iforest import IForest
from pyod.models.knn import KNN
from pyod.models.lof import LOF
from pyod.models.mcd import MCD
from pyod.models.ocsvm import OCSVM
from pyod.models.pca import PCA

from pyod.utils.utility import standardizer
from pyod.utils.utility import precision_n_scores
from sklearn.metrics import roc_auc_score

from DL import DL
from pyod.models.abdl import ABDL
from pyod.models.kabdl import KABDL
from pyod.models.ekabdl import EKABDL
from pyod.models.kabod import KABOD
from pyod.models.ekabod import EKABOD

# TODO: add neural networks, LOCI, SOS, COF, SOD

n_components = 50
n_nonzero_coefs = 3
fit_algorithm = 'aksvd'
transform_algorithm = 'omp'
max_iter = 10
n_neighbors = 5

# Define data file and read X and y
mat_file_list = [
    'arrhythmia.mat',
    'cardio.mat',
    'glass.mat',
    'ionosphere.mat',
    'letter.mat',
    'lympho.mat',
    'mnist.mat',
    'musk.mat',
    'optdigits.mat',
    'pendigits.mat',
    'pima.mat',
    'satellite.mat',
    'satimage-2.mat',
    'shuttle.mat',
    'vertebral.mat',
    'vowels.mat',
    'wbc.mat'
]

# define the number of iterations
n_ite = 1
n_classifiers = 16

df_columns = ['Data', '#Samples', '# Dimensions', 'Outlier Perc',
              'CBLOF', 'FB', 'HBOS', 'IForest', 'KNN', 'LOF', 'MCD', 'OCSVM', 'PCA',
              'ABOD', 'KABOD', 'EKABOD',
              'DL', 'ABDL', 'KABDL', 'EKABDL'
]

# initialize the container for saving the results
roc_df = pd.DataFrame(columns=df_columns)
prn_df = pd.DataFrame(columns=df_columns)
time_df = pd.DataFrame(columns=df_columns)

for j in range(len(mat_file_list)):

    mat_file = mat_file_list[j]
    mat = loadmat(os.path.join('data', mat_file))

    X = mat['X']
    y = mat['y'].ravel()
    outliers_fraction = np.count_nonzero(y) / len(y)
    outliers_percentage = round(outliers_fraction * 100, ndigits=4)

    # compute the number of atoms
    n_features = X.shape[1]

    # construct containers for saving results
    roc_list = [mat_file[:-4], X.shape[0], X.shape[1], outliers_percentage]
    prn_list = [mat_file[:-4], X.shape[0], X.shape[1], outliers_percentage]
    time_list = [mat_file[:-4], X.shape[0], X.shape[1], outliers_percentage]

    roc_mat = np.zeros([n_ite, n_classifiers])
    prn_mat = np.zeros([n_ite, n_classifiers])
    time_mat = np.zeros([n_ite, n_classifiers])

    for i in range(n_ite):
        print("\n... Processing", mat_file, '...', 'Iteration', i + 1)
        random_state = np.random.RandomState(i)

        # 60% data for training and 40% for testing
        X_train, X_test, y_train, y_test = \
            train_test_split(X, y, test_size=0.4, random_state=random_state)

        # standardizing data for processing
        X_train_norm, X_test_norm = standardizer(X_train, X_test)

        classifiers = {
            'Cluster-based Local Outlier Factor': CBLOF(
                n_clusters=10,
                contamination=outliers_fraction,
                check_estimator=False,
                random_state=random_state),
            'Feature Bagging': FeatureBagging(contamination=outliers_fraction,
                                              random_state=random_state),
            'Histogram-base Outlier Detection (HBOS)': HBOS(
                contamination=outliers_fraction),
            'Isolation Forest': IForest(contamination=outliers_fraction,
                                        random_state=random_state),
            'K Nearest Neighbors (KNN)': KNN(contamination=outliers_fraction),
            'Local Outlier Factor (LOF)': LOF(
                contamination=outliers_fraction),
            'Minimum Covariance Determinant (MCD)': MCD(
                contamination=outliers_fraction, random_state=random_state),
            'One-class SVM (OCSVM)': OCSVM(contamination=outliers_fraction),
            'Principal Component Analysis (PCA)': PCA(
                contamination=outliers_fraction, random_state=random_state),
            'Angle-based Outlier Detector (ABOD)': ABOD(
                contamination=outliers_fraction, n_neighbors=n_neighbors),
            'Kernel Angle-based Outlier Detector (KABOD)': KABOD(
                contamination=outliers_fraction, n_neighbors=n_neighbors),
            'Extended Kernel Angle-based Outlier Detector (EKABOD)': EKABOD(
                contamination=outliers_fraction, n_neighbors=n_neighbors),
            'Dictionary Learning (DL)': DL(
                n_components=n_components,n_nonzero_coefs=n_nonzero_coefs,
                fit_algorithm=fit_algorithm, transform_algorithm=transform_algorithm,
                max_iter=max_iter, contamination=outliers_fraction,
                random_state=random_state),
            'Angle-based Dictionary Learning (ABDL)': ABDL(
                n_components=n_components, n_nonzero_coefs=n_nonzero_coefs,
                fit_algorithm=fit_algorithm, transform_algorithm=transform_algorithm,
                max_iter=max_iter, contamination=outliers_fraction,
                n_neighbors=n_neighbors, random_state=random_state),
            'Kernel Angle-based Dictionary Learning (KABDL)': KABDL(
                n_components=n_components, n_nonzero_coefs=n_nonzero_coefs,
                fit_algorithm=fit_algorithm, transform_algorithm=transform_algorithm,
                max_iter=max_iter, contamination=outliers_fraction,
                n_neighbors=n_neighbors, random_state=random_state),
            'Extended Kernel Angle-based Dictionary Learning (EKABDL)': EKABDL(
                n_components=n_components, n_nonzero_coefs=n_nonzero_coefs,
                fit_algorithm=fit_algorithm, transform_algorithm=transform_algorithm,
                max_iter=max_iter, contamination=outliers_fraction,
                n_neighbors=n_neighbors, random_state=random_state),
        }
        classifiers_indices = {
            'Cluster-based Local Outlier Factor': 0,
            'Feature Bagging': 1,
            'Histogram-base Outlier Detection (HBOS)': 2,
            'Isolation Forest': 3,
            'K Nearest Neighbors (KNN)': 4,
            'Local Outlier Factor (LOF)': 5,
            'Minimum Covariance Determinant (MCD)': 6,
            'One-class SVM (OCSVM)': 7,
            'Principal Component Analysis (PCA)': 8,
            'Angle-based Outlier Detector (ABOD)': 9,
            'Kernel Angle-based Outlier Detector (KABOD)': 10,
            'Extended Kernel Angle-based Outlier Detector (EKABOD)': 11,
            'Dictionary Learning (DL)': 12,
            'Angle-based Dictionary Learning (ABDL)': 13,
            'Kernel Angle-based Dictionary Learning (KABDL)': 14,
            'Extended Kernel Angle-based Dictionary Learning (EKABDL)': 15,
        }

        for clf_name, clf in classifiers.items():
            t0 = time()
            clf.fit(X_train_norm)
            test_scores = clf.decision_function(X_test_norm)
            t1 = time()
            duration = round(t1 - t0, ndigits=4)

            roc = round(roc_auc_score(y_test, test_scores), ndigits=4)
            prn = round(precision_n_scores(y_test, test_scores), ndigits=4)

            print('{clf_name} ROC:{roc}, precision @ rank n:{prn}, '
                  'execution time: {duration}s'.format(
                clf_name=clf_name, roc=roc, prn=prn, duration=duration))

            time_mat[i, classifiers_indices[clf_name]] = duration
            roc_mat[i, classifiers_indices[clf_name]] = roc
            prn_mat[i, classifiers_indices[clf_name]] = prn

    time_list = time_list + np.mean(time_mat, axis=0).tolist()
    temp_df = pd.DataFrame(time_list).transpose()
    temp_df.columns = df_columns
    time_df = pd.concat([time_df, temp_df], axis=0)

    roc_list = roc_list + np.mean(roc_mat, axis=0).tolist()
    temp_df = pd.DataFrame(roc_list).transpose()
    temp_df.columns = df_columns
    roc_df = pd.concat([roc_df, temp_df], axis=0)

    prn_list = prn_list + np.mean(prn_mat, axis=0).tolist()
    temp_df = pd.DataFrame(prn_list).transpose()
    temp_df.columns = df_columns
    prn_df = pd.concat([prn_df, temp_df], axis=0)

    properties = '_rbf_n_components_{}_n_nonzero_coefs_{}_max_iter_{}_n_neighbors_{}_'.format(n_components, n_nonzero_coefs, max_iter, n_neighbors)

    # Save the results for each run
    time_df.to_csv('abod_time' + properties + '.csv', index=False, float_format='%.3f')
    roc_df.to_csv('abod_roc' + properties + '.csv', index=False, float_format='%.3f')
    prn_df.to_csv('abod_prc' + properties + '.csv', index=False, float_format='%.3f')
